/*
 * Image plugin to VDR (C++)
 *
 * (C) 2003 Kai Tobias Burwieck <kai at burwieck.net>
 * (C) 2004     "Interpohl"  <interpohl at vdr-portal.de> 
 * (C) 2004     A. Brachold   <anbr at users.berlios.de>
 * (C) 2004     O. Kreuzinger <Onno at Kreuzinger.biz>
 * (C) 2004     A. Holzhammer for the massive script updates
 *  
 *  based on mp3/mplayer plguin by Stefan Hlswitt <huels at iname.com>
 *   
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <getopt.h>
#include <vdr/plugin.h>
#include <vdr/remote.h>

#include "setup-image.h"
#include "data-image.h"
#include "control-image.h"
#include "config.h"
// #include "i18n.h"

static const char *VERSION        = "1.99.0";
static const char *DESCRIPTION    = trNOOP("An Image Viewer plugin for albumcover");
static const char *MAINMENUENTRY  = trNOOP("Coverviewer");

class cPluginImage : public cPlugin {
private:
#if APIVERSNUM >= 10330
  bool ExternalPlay(const char *path, bool test);
#endif
public:
  virtual ~cPluginImage();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return tr(DESCRIPTION); }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Start(void);
  virtual const char *MainMenuEntry(void) { return tr(MAINMENUENTRY); }
  virtual cOsdObject *MainMenuAction(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
#if APIVERSNUM >= 10331
  virtual const char **SVDRPHelpPages(void);
  virtual cString SVDRPCommand(const char *Command, const char *Option, int &ReplyCode);
#endif
};

bool cPluginImage::SetupParse(const char *szName, const char *szValue)
{
	return ImageSetup.SetupParse(szName,szValue);
}

const char *cPluginImage::CommandLineHelp(void)
{
  static char *help_str=0;
  
  free(help_str);    //                                     for easier orientation, this is column 80|
  asprintf(&help_str,"  -C PATH,  --config=PATH  to specify directory of configuration data\n"
                     "                           to the directory 'moronsuite'\n"
                     "                           (default: \"%s\")\n",
                    configdir?configdir:""
                    );
  return help_str;
}

bool cPluginImage::ProcessArgs(int argc, char *argv[])
{
  static struct option long_options[] = {
      { "config",   required_argument, NULL, 'C' },
      { NULL }
    };
    
  int c, option_index = 0;
  while((c=getopt_long(argc,argv,"c:",long_options,&option_index))!=-1) {
    switch (c) {
      case 'C': configdir=optarg;break;
      default:  return false;
      }
    }
  return true;
}

bool cPluginImage::Start(void)
{
  if(configdir == NULL) {
    configdir = strdup(ConfigDirectory(NULL));
    isyslog("coverviewer: --config not set, use default path '%s'", configdir);
    }
  else
    isyslog("coverviewer: --config set: '%s'", configdir);



  ImageSources.Load(AddDirectory(configdir,  "moronsuite/coverviewer/coversources.conf"));
  if(ImageSources.Count()<1) {
    esyslog("coverviewer: you must have defined at least one source in coversources.conf");
    return false;
    }
  
//   RegisterI18n(Phrases);
  return true;
}

cPluginImage::~cPluginImage()
{
}


cOsdObject *cPluginImage::MainMenuAction(void)
{
  return new cCoverControl;
}


bool cPluginImage::ExternalPlay(const char *path, bool test)
{
  char real[PATH_MAX+1];

  if(realpath(path,real)) {

    cFileSource *src=ImageSources.FindSource(real, true);
    if(src) {
      cDirItem *item=new cDirItem(src, 0, 0, itDir); 
      if(item) {
        item->SplitAndSet(real);
        if(item->GuessType()) {
          cSlideShow *newss = new cSlideShow(item);
          if(newss && newss->Load() && newss->Count()) {
            if(!test) {
              cCoverControl::SetSlideShow(newss);
              cRemote::CallPlugin("coverviewer");
              }
            else
              delete newss;
            delete item;
            return true;
            }
          else
            dsyslog("coverviewer: service: error building playlist");
          delete newss;
          }
        else
          dsyslog("coverviewer: service: GuessType() failed for '%s'", path);
        delete item;
        }
      }
    else dsyslog("coverviewer: service: cannot find source for '%s', real '%s'", path, real);
    }
  else
    if(errno!=ENOENT && errno!=ENOTDIR)
      esyslog("coverviewer: ERROR: realpath: %s: %s", path, strerror(errno));

  return false;
}

const char **cPluginImage::SVDRPHelpPages(void)
{
  static const char *HelpPages[] = {
    "PLAYDIR <dirname>\n"
    "    Triggers playback of directory 'dirname'.",
    NULL
    };
  return HelpPages;
}

cString cPluginImage::SVDRPCommand(const char *Command, const char *Option, int &ReplyCode)
{
  if(!strcasecmp(Command,"PLAYDIR")) {

    if(*Option) {
      if(ExternalPlay(Option,false))
        return "Playback triggered";
      else {
        ReplyCode=550;
        return "Playback failed";
       }
    }
    else {
      ReplyCode=501;
      return "Missing dirname";
    }


  }

  return NULL;
}


cMenuSetupPage *cPluginImage::SetupMenu(void)
{
  return new cMenuSetupImage;
}

VDRPLUGINCREATOR(cPluginImage); // Don't touch this!
