///
///	@file dia.c		@brief diashow module
///
///	Copyright (c) 2009 - 2012 by Johns.  All Rights Reserved.
///
///	Contributor(s):
///
///	License: AGPLv3
///
///	This program is free software: you can redistribute it and/or modify
///	it under the terms of the GNU Affero General Public License as
///	published by the Free Software Foundation, either version 3 of the
///	License.
///
///	This program is distributed in the hope that it will be useful,
///	but WITHOUT ANY WARRANTY; without even the implied warranty of
///	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
///	GNU Affero General Public License for more details.
///
///	$Id: $
//////////////////////////////////////////////////////////////////////////////

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <dirent.h>

#include <libintl.h>
#define _(str) gettext(str)		///< gettext shortcut
#define _N(str) str			///< gettext_noop shortcut

#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <pthread.h>

#ifndef USE_JPEG
#define USE_JPEG			///< support jpeg images
#endif
#ifndef USE_PNG
#define USE_PNG				///< support png images
#endif
#ifndef USE_ZZIPLIB
#undef USE_ZZIPLIB			///< zip archives support
#endif
#ifndef USE_AVFS
#undef USE_AVFS				///< a virtual file system support
#endif
#ifndef USE_SWSCALE
#define USE_SWSCALE			///< support ffmpeg swscale
#endif

#ifdef USE_ZZIPLIB
#include <zzip/lib.h>
#endif
#ifdef USE_AVFS
#include <virtual.h>
#endif
#ifdef USE_SWSCALE
#include <libswscale/swscale.h>
#endif

#include "misc.h"
#include "readdir.h"
#include "dia.h"

//7///////////////////////////////////////////////////////////////////////////
//	Image
//////////////////////////////////////////////////////////////////////////////

/**
**	Image structure.
*/
typedef struct _image_ Image;

/**
**	Image structure.
*/
struct _image_
{
    uint16_t Width;			///< image width
    uint16_t Height;			///< image height
    uint8_t Data[1];			///< image data 32bit ARGB
};

/**
**	Rotate image.
*/
enum _rotate_
{
    RotateNone,				///< don't rotate
    Rotate90CW,				///< 90 rotate
    Rotate180CW,			///< 180 rotate
    Rotate90CCW,			///< -90 rotate
};

/**
**	Scaler method.
*/
enum _scale_method_
{
    ScaleFastBilinear,			///< fast bilinear
    ScaleBilinear,			///< bilinear
    ScaleBicubic,			///< bicubic
    ScaleX,				///< X
    ScalePoint,				///< point
    ScaleArea,				///< area
    ScaleBicubLin,			///< bicubic linear
    ScaleGauss,				///< gauss
    ScaleSinc,				///< sinc
    ScaleLanczos,			///< lanczos
    ScaleSpline,			///< spline
};

//////////////////////////////////////////////////////////////////////////////

/**
**	Rotate an image 90 clockwise.
**
**	@param src	image to rotate clockwise
**	@param dst	where to put the rotated data
*/
void ImageRotate90(const Image * src, Image * dst)
{
    const uint8_t *base;
    const uint8_t *endp;
    uint8_t *data;
    int width;
    int height;

    data = dst->Data;
    width = src->Width * sizeof(uint32_t);
    height = src->Height;
    endp = src->Data + width * height;
    for (base = endp - width; base < endp; base += sizeof(uint32_t)) {
	int i;
	const uint8_t *tmp;

	for (i = 0, tmp = base; i < height; i++, tmp -= width) {
	    *data++ = tmp[0];
	    *data++ = tmp[1];
	    *data++ = tmp[2];
	    *data++ = tmp[3];
	}
    }
}

/**
**	Rotate an image.
*/
VdrImage *ImageRotate(VdrImage * image, int rotate)
{
    switch (rotate) {
	case RotateNone:
	    return image;
    }
    return image;
}

/**
**	Scale an image.
*/
VdrImage * ImageScale(VdrImage * image, int width, int height, int method)
{
#ifdef USE_SWSCALE
    struct SwsContext *sws_context;
    int src_width;
    int src_height;
    int sws_flags;
    const uint8_t *src_data;
    int src_linesize;
    VdrImage * dst_image;
    uint8_t *dst_data;
    int dst_linesize;

    src_width = VdrImageWidth(image);
    src_height = VdrImageHeight(image);
    if ( src_width == width && src_height == height ) {
	return image;			// correct size, nothing todo
    }
    src_data = VdrImageData(image);
    src_linesize = src_width * sizeof(uint32_t);

    if (!(dst_image = VdrImageNew(width, height, NULL))) {
	Error(_("dia: can't create new image\n"));
	return image;
    }
    dst_data = VdrImageData(dst_image);
    dst_linesize = width * sizeof(uint32_t);

    sws_context = NULL;			// FIXME: try to reuse
    switch (method) {
	case ScaleFastBilinear:
	default:
	    sws_flags = SWS_FAST_BILINEAR;
	    break;
	case ScaleBilinear:
	    sws_flags = SWS_BILINEAR;
	    break;
	case ScaleBicubic:
	    sws_flags = SWS_BICUBIC;
	    break;
	case ScaleX:
	    sws_flags = SWS_X;
	    break;
	case ScalePoint:
	    sws_flags = SWS_POINT;
	    break;
	case ScaleArea:
	    sws_flags = SWS_AREA;
	    break;
	case ScaleBicubLin:
	    sws_flags = SWS_BICUBLIN;
	    break;
	case ScaleGauss:
	    sws_flags = SWS_GAUSS;
	    break;
	case ScaleSinc:
	    sws_flags = SWS_SINC;
	    break;
	case ScaleLanczos:
	    sws_flags = SWS_LANCZOS;
	    break;
	case ScaleSpline:
	    sws_flags = SWS_SPLINE;
	    break;
    }
    printf("%x\n", sws_flags);
    sws_context =
	sws_getCachedContext(sws_context, src_width, src_height, PIX_FMT_ARGB,
	width, height, PIX_FMT_ARGB, sws_flags, NULL, NULL, NULL);
    if( !sws_context ) {
	Error(_("dia: can't create scale context\n"));
	VdrImageDel(dst_image);
	return image;
    }

    sws_scale(sws_context, &src_data, &src_linesize, 0, src_height, &dst_data,
	&dst_linesize);

    sws_freeContext(sws_context);
    VdrImageDel(image);
    return dst_image;
#else
    return image;
#endif
}

//////////////////////////////////////////////////////////////////////////////
//	JPEG
//////////////////////////////////////////////////////////////////////////////

#ifdef USE_JPEG				// {

#include <setjmp.h>
#include <string.h>
#include <errno.h>
#include <jpeglib.h>
#include <endian.h>

    // funny construct to have noreturn attribute with stupid gcc
static void ImageJPEGErrorExit(j_common_ptr cinfo) __attribute__ ((noreturn));

/**
**	Called from jpeg on errors by error_exit method.
**
**	@param cinfo	JPEG common info structure
*/
static void ImageJPEGErrorExit(j_common_ptr cinfo)
{
    longjmp(cinfo->client_data, 1);
}

/**
**	Load a JPEG image from given file name.
**
**	@param name	file name to open JPEG file
**	@param width	requested output size (or 0)
**	@param height	requested output size (or 0)
**
**	@returns loaded BGRA/ARGB image, NULL if failure.
**
**	@see file:///usr/share/doc/jpeg-6b-r8/libjpeg.doc.bz2
**	@see file:///usr/share/doc/jpeg-7/libjpeg.txt.bz2
*/
VdrImage *ImageLoadJPEG0(const char *name, unsigned width, unsigned height)
{
    VdrImage *image;
    jmp_buf jmpbuf;
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    JSAMPARRAY buffer;
    uint8_t *data;

#ifdef USE_AVFS
    int fildes;
    struct stat stat_buf;
    uint8_t *in_buf;
    ssize_t in_size;

    //
    //	read complete file into buffer
    //
    if ((fildes = virt_open(name, O_RDONLY, 0)) < 0) {
	Debug(3, "%s: can't open %s: %s\n", __FUNCTION__, name,
	    strerror(errno));
	return NULL;
    }
    if (virt_fstat(fildes, &stat_buf) < 0) {
	Error("diashow: can't stat file '%s': %s\n", name, strerror(errno));
	virt_close(fildes);
	return NULL;
    }
    in_size = stat_buf.st_size;

    if (!(in_buf = malloc(in_size))) {
	Error("diashow: out of memory file '%s': %s\n", name, strerror(errno));
	virt_close(fildes);
	return NULL;
    }
    if (virt_read(fildes, in_buf, in_size) != in_size) {
	Error("diashow: can't read file '%s': %s\n", name, strerror(errno));
	free(in_buf);
	virt_close(fildes);
	return NULL;
    }
    virt_close(fildes);

#else
    FILE *fd;

    // FIXME: rewrite to use virt_open

    // open the file
    if (!(fd = fopen(name, "rb"))) {
	Debug(3, "%s: can't open %s: %s\n", __FUNCTION__, name,
	    strerror(errno));
	return NULL;
    }
#endif

    image = NULL;
    if (setjmp(jmpbuf)) {		// return here, if any errors
	if (image) {
	    VdrImageDel(image);
	}

	jpeg_destroy_decompress(&cinfo);
#ifdef USE_AVFS
	free(in_buf);
#else
	fclose(fd);
#endif

	return NULL;
    }
    // setup the error handler
    cinfo.err = jpeg_std_error(&jerr);
    jerr.error_exit = ImageJPEGErrorExit;

    // prepare to load the file
    jpeg_create_decompress(&cinfo);
    cinfo.client_data = jmpbuf;
#ifdef USE_AVFS
    jpeg_mem_src(&cinfo, in_buf, in_size);
#else
    jpeg_stdio_src(&cinfo, fd);
#endif

    // check the header
    jpeg_read_header(&cinfo, TRUE);

    // setup decoder for greater speed
    cinfo.do_fancy_upsampling = FALSE;
    cinfo.do_block_smoothing = FALSE;
    cinfo.dct_method = JDCT_IFAST;
    cinfo.dither_mode = JDITHER_ORDERED;

    //
    //	prescale to requested output size
    //
    if (width || height) {
	int scale_w;
	int scale_h;

	// jpeg docs says: n 1/8 .. 16/8 are supported
	scale_w = (width * cinfo.scale_denom) / cinfo.image_width;
	scale_h = (height * cinfo.scale_denom) / cinfo.image_height;

	cinfo.scale_num = scale_w;
	if (scale_h > scale_w) {
	    cinfo.scale_num = scale_h;
	}
	if (cinfo.scale_num >= 16) {
	    cinfo.scale_num = 16;
	} else if (cinfo.scale_num < 1) {
	    cinfo.scale_num = 1;
	}
#if 0
	cinfo.scale_num = 1;

	scale_w = cinfo.image_width / width;
	scale_h = cinfo.image_height / height;

	cinfo.scale_denom = scale_w;
	if (scale_h < scale_w) {
	    cinfo.scale_denom = scale_h;
	}
	if (cinfo.scale_denom >= 8) {
	    cinfo.scale_denom = 8;
	} else if (cinfo.scale_denom < 1) {
	    cinfo.scale_denom = 1;
	} else if (cinfo.scale_denom == 3) {
	    cinfo.scale_denom = 2;
	} else if (cinfo.scale_denom == 5) {
	    cinfo.scale_denom = 4;
	} else if (cinfo.scale_denom == 6) {
	    cinfo.scale_denom = 4;
	} else if (cinfo.scale_denom == 7) {
	    cinfo.scale_denom = 4;
	}
#endif
    }
    jpeg_calc_output_dimensions(&cinfo);

    // start decompression
    jpeg_start_decompress(&cinfo);

    // dimension is known after jpeg_start_decompress
    buffer =
	cinfo.mem->alloc_sarray((j_common_ptr) & cinfo, JPOOL_IMAGE,
	cinfo.output_width * cinfo.output_components, cinfo.rec_outbuf_height);
    if (!(image = VdrImageNew(cinfo.output_width, cinfo.output_height, NULL))) {
	fprintf(stderr, "dia: can't create new image\n");

	// clean up
	jpeg_destroy_decompress(&cinfo);
#ifdef USE_AVFS
	free(in_buf);
#else
	fclose(fd);
#endif
	return NULL;
    }
    data = VdrImageData(image);

    // FIXME: loading can be done quicker, f.e. more line at once...

    // read lines
    switch (cinfo.output_components) {
	case 1:			// grayscale
	    while (cinfo.output_scanline < cinfo.output_height) {
		unsigned x;

		jpeg_read_scanlines(&cinfo, buffer, 1);
		for (x = 0; x < cinfo.output_width; x++) {
#if BYTE_ORDER == BIG_ENDIAN
		    *data++ = 0xFF;	// alpha
		    *data++ = GETJSAMPLE(buffer[0][x]);
		    *data++ = GETJSAMPLE(buffer[0][x]);
		    *data++ = GETJSAMPLE(buffer[0][x]);
#else
		    *data++ = GETJSAMPLE(buffer[0][x]);
		    *data++ = GETJSAMPLE(buffer[0][x]);
		    *data++ = GETJSAMPLE(buffer[0][x]);
		    *data++ = 0xFF;	// alpha
#endif
		}
	    }
	    break;
	default:			// RGB
	    while (cinfo.output_scanline < cinfo.output_height) {
		unsigned x;

		jpeg_read_scanlines(&cinfo, buffer, 1);
		for (x = 0; x < cinfo.output_width; x++) {
#if BYTE_ORDER == BIG_ENDIAN
		    *data++ = 0xFF;	// alpha
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 0]);
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 1]);
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 2]);
#else
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 2]);
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 1]);
		    *data++ = GETJSAMPLE(buffer[0][x * 3 + 0]);
		    *data++ = 0xFF;	// alpha
#endif
		}
	    }
	    break;
    }

    // clean up
    jpeg_destroy_decompress(&cinfo);
#ifdef USE_AVFS
    free(in_buf);
#else
    fclose(fd);
#endif

    return image;
}

/**
**	Load a JPEG image from given file name.
**
**	@param name	file name to open JPEG file
**
**	@returns loaded BGRA/ARGB image, NULL if failure.
*/
static VdrImage *ImageLoadJPEG(const char *name)
{
    return ImageLoadJPEG0(name, 0, 0);
}

#else // }{ USE_JPEG

    /// Dummy for load a JPEG image from given file name.
#define ImageLoadJPEG(name)	NULL

#endif // } USE_JPEG

//////////////////////////////////////////////////////////////////////////////
//	FileEntry / Filelist
//////////////////////////////////////////////////////////////////////////////

/**
**	File list entry typedef.
*/
typedef struct _file_entry_ FileEntry;

/**
**	File list entry structure.
*/
struct _file_entry_
{
    char *Name;				///< name
};

/**
**	File list typedef.
*/
typedef struct _file_list_ FileList;

/**
**	File list structure.
*/
struct _file_list_
{
    int Size;				///< number of elements in ::Entry
    FileEntry Entry[1];			///< table of file name entries
};

/**
**	Get entry at index in list.
**
**	@param filelist	a list of file names
**	@param index	integer index into list
*/
static const FileEntry *EntryGet(const FileList * filelist, int index)
{
    if (index < filelist->Size) {
	return &filelist->Entry[index];
    }
    return NULL;
}

/**
**	First index/entry in list.
**
**	@param filelist		a list of file names
**	@param[in,out] index	integer index into list
*/
const FileEntry *EntryFirst(const FileList * filelist, int *index)
{
    if (*index < 0) {
	*index = 0;
    }
    if (*index < filelist->Size) {
	return &filelist->Entry[*index];
    }
    return NULL;
}

/**
**	Next index/entry in list.
**
**	@param filelist		a list of file names
**	@param[in,out] index	integer index into list
*/
const FileEntry *EntryNext(const FileList * filelist, int *index)
{
    if (*index < 0) {
	*index = -1;
    }
    if (*index < filelist->Size - 1) {
	++*index;
	return &filelist->Entry[*index];
    }
    return NULL;
}

/**
**	Prev index/entry in list.
**
**	@param filelist		a list of file names
**	@param[in,out] index	integer index into list
*/
const FileEntry *EntryPrev(const FileList * filelist, int *index)
{
    if (*index > filelist->Size - 1) {
	*index = filelist->Size;
    }
    if (*index > 0) {
	--*index;
	return &filelist->Entry[*index];
    }
    return NULL;
}

/**
**	Get file entry name.
**
**	@param entry	file name list entry
*/
const char *EntryName(const FileEntry * entry)
{
    return entry->Name;
}

//////////////////////////////////////////////////////////////////////////////

/**
**	Table of supported image suffixes.
*/
static const NameFilter ImageFilters[] = {
#define FILTER(x) { sizeof(x) - 1, x }
    FILTER(".jpg"),
    FILTER(".png"),
#undef FILTER
    {0, NULL}
};

/**
**	Read directory in filelist.
**
**	@param name	directory name
**
**	@returns filelist with only image file names.
*/
static FileList *FileListNew(const char *name)
{
    char **namelist;
    FileList *filelist;
    int n;
    int i;

    n = ScanDirectory(name, 0, ImageFilters, &namelist);
    if (n >= 0) {
	filelist = malloc(sizeof(FileList) + n * sizeof(FileEntry));
	if (filelist) {
	    filelist->Size = n;
	    for (i = 0; i < n; ++i) {
		filelist->Entry[i].Name = strdup(namelist[i]);
		free(namelist[i]);
	    }
	    free(namelist);
	}
	return filelist;
    }
    return NULL;
}

/**
**	Free directory filelist.
**
**	@param filelist	filelist with malloced strings
*/
static void FileListDel(FileList * filelist)
{
    int index;

    if (filelist) {
	for (index = 0; index < filelist->Size; ++index) {
	    free(filelist->Entry[index].Name);
	}
	free(filelist);
    }
}

//////////////////////////////////////////////////////////////////////////////
//	Diashow
//////////////////////////////////////////////////////////////////////////////

    /// default film-strip thumbnail width
#define DIA_FILM_STRIP_WIDTH 170
    /// default film-strip thumbnail height
#define DIA_FILM_STRIP_HEIGHT 96

    /// default index thumbnail width
#define DIA_INDEX_WIDTH (170 * 2)
    /// default index thumbnail height
#define DIA_INDEX_HEIGHT (96 * 2)

    /// default corner command width
#define DIA_CORNER_WIDTH (170 / 2)
    /// default corner command height
#define DIA_CORNER_HEIGHT (96 / 2)

    /// min macro
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
    /// max macro
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))

/**
**	Dia display layout mode.
*/
typedef enum _dia_layout_
{
    DIA_LAYOUT_INDEX,			///< index-page
    DIA_LAYOUT_SINGLE,			///< single picture
    // DIA_LAYOUT_DOUBLE,		///< double picture
} DiaLayout;

/**
**	Dia global variables structure typedef.
*/
typedef struct _dia_ Dia;

/**
**	Dia global variables structure.
*/
struct _dia_
{
    uint16_t Width;			///< window width
    uint16_t Height;			///< window height

    uint16_t CornerWidth;		///< corner command area width
    uint16_t CornerHeight;		///< corner command area height

    uint16_t FilmStripWidth;		///< film-strip thumbnail width
    uint16_t FilmStripHeight;		///< film-strip thumbnail height
    uint16_t FilmStripScale;		///< FIXME: film-strip selected scale

    uint16_t IndexWidth;		///< index-page thumbnail width
    uint16_t IndexHeight;		///< index-page thumbnail height

    uint16_t AspectNum;			///< thumbnail aspect ratio
    uint16_t AspectDen;			///< thumbnail aspect ratio

    uint16_t SlideShowDelay;		///< slide show delay

    unsigned DoneInit:1;		///< init done
    unsigned NeedRedraw:1;		///< need to redraw
    unsigned SlideShow:1;		///< slide show running
    unsigned IndexLabel:1;		///< print file names on thumbnails
    unsigned ImageLabel:1;		///< print file names on image

    DiaLayout Layout:1;			///< dia window layout
    int FilmStrip:1;			///< flag show film-strip
    int FilmStripHorizontal:1;		///< flag use horizontal film-strip
    int Backdrop:1;			///< flag become desktop background
    int Fullscreen:1;			///< flag become fullscreen

    uint32_t SlideShowTick;		///< last slide-show tick

    uint32_t LastTime;			///< last handled event time
    int16_t LastX;			///< last x-coordinate
    int16_t LastY;			///< last y-coordinate
    int State;				///< state for pointer buttons handle

    int16_t OffsetX;			///< offset for slide effect
    int16_t OffsetY;			///< offset for slide effect

    // Array *DirTree;			///< directory tree
    char *Path;				///< current path
    FileList *FilesInDir;		///< files in directory
    int CurrentIndex;			///< current file index
    int FirstIndex;			///< first index of film-strip/index

    int Rotate;				///< image orientation
    int ScaleMethod;			///< scaler method

    // DiaCache Cache[DIA_CACHE_SIZE];	///< image cache

    struct _colors_
    {
	uint32_t Background;		///< background color
	uint32_t Highlight;		///< highlighted color
	uint32_t Shadow;		///< text shadow
	uint32_t Text;			///< text foreground
    } Color[1];				///< predefined color
};

static Dia DiaVars[1];			///< diashow globals

//////////////////////////////////////////////////////////////////////////////
//	Functions
//////////////////////////////////////////////////////////////////////////////

/**
**	Show image.
**
**	@param name	file name of image to draw
**	@param x	x-offset on drawable to render image
**	@param y	y-offset on drawable to render image
**	@param width	width of image to display
**	@param height	height of image to display
**
**	@todo cache images.
*/
static void DiaShowImage(const char *name, int x, int y, unsigned width,
    unsigned height)
{
    char *buf;
    VdrImage *image;

    // FIXME: cache images
    buf = alloca(strlen(name) + strlen(DiaVars->Path) + 2);
    stpcpy(stpcpy(stpcpy(buf, DiaVars->Path), "/"), name);
    // ImageLoad()
    if ((image = ImageLoadJPEG0(buf, width, height))) {
	int i;
	unsigned w;
	unsigned h;

	// keep image aspect ratio
	i = (VdrImageWidth(image) * 65536) / VdrImageHeight(image);
	w = MIN(width * 65536, height * i);
	h = MIN(height, w / i);
	w = (h * i) / 65536;
	if (w < 1) {
	    w = 1;
	}
	if (h < 1) {
	    h = 1;
	}
	// pre scale the image
	image = ImageScale(image, w, h, DiaVars->ScaleMethod);
	// center image inside the box
	DrawImage(image, x + (width - w) / 2, y + (height - h) / 2, w, h);
	VdrImageDel(image);
    }
}

/**
**	Show thumbnail.
**
**	Short cut for thumbnails
**
**	@param name	file name of thumnail to draw
**	@param x	x-offset on drawable to thumnail image
**	@param y	y-offset on drawable to thumnail image
**	@param width	width of thumnail to display
**	@param height	height of thumnail to display
**
**	@todo cache thumbnails
*/
static void DiaShowThumbnail(const char *name, int x, int y, unsigned width,
    unsigned height)
{
    VdrImage *image;
    char *buf;

    buf = alloca(strlen(name) + strlen(DiaVars->Path) + 2);
    stpcpy(stpcpy(stpcpy(buf, DiaVars->Path), "/"), name);
    if ((image = ImageLoadJPEG0(buf, width, height))) {
	int i;
	unsigned w;
	unsigned h;

	// keep image aspect ratio
	i = (VdrImageWidth(image) * 65536) / VdrImageHeight(image);
	w = MIN(width * 65536, height * i);
	h = MIN(height, w / i);
	w = (h * i) / 65536;
	if (w < 1) {
	    w = 1;
	}
	if (h < 1) {
	    h = 1;
	}
	// center thumbnail inside the box
	DrawImage(image, x + (width - w) / 2, y + (height - h) / 2, w, h);
	VdrImageDel(image);
    } else {
	DiaShowImage(name, x, y, width, height);
    }
}

/**
**	Draw thumbnail page.
**
**	@param x0	x start coordinate of thumbnail array
**	@param y0	y start coordinate of thumbnail array
**	@param w0	width in pixel of thumbnail array
**	@param h0	height in pixel of thumbnail array
**	@param tw	thumbnail width
**	@param th	thumbnail height
**	@oaram tb	thumbnail border
*/
static void DiaDrawPage(int x0, int y0, int w0, int h0, int tw, int th, int tb)
{
    int x;
    int y;
    int index;
    const FileEntry *entry;

    index = DiaVars->FirstIndex;
    entry = EntryFirst(DiaVars->FilesInDir, &index);
    y = y0;
    do {
	x = x0;
	do {
	    if (entry) {
		const char *file;

		file = EntryName(entry);
		if (!file) {
		    break;
		}
		// highlight current file
		if (index == DiaVars->CurrentIndex) {
		    DrawRectangle(x, y, x + tw, y + th,
			DiaVars->Color->Highlight);
		}
		// FIXME: show current file bigger
		DiaShowThumbnail(file, x + tb, y + tb, tw - 2 * tb,
		    th - 2 * tb);
		if (DiaVars->IndexLabel) {
		    DrawText(x + 2, y + 2, file, DiaVars->Color->Shadow, 0x0UL,
			tw - 3, 0, 4 + 1);
		    DrawText(x + 1, y + 1, file, DiaVars->Color->Text, 0x0UL,
			tw - 4, 0, 4 + 1);
		}

		entry = EntryNext(DiaVars->FilesInDir, &index);
	    }
	    x += tw;
	} while (x <= w0 - tw);
	y += th;
    } while (y <= h0 - th);
}

/**
**	Draw vertical film-strip.
**
**	@param x0	film-strip x start coordinate
**	@param y0	film-strip y start coordinate
*/
static void DiaDrawVerticalFilmStrip(int x0, int y0)
{
    // FIXME: double page film-strip
    DiaDrawPage(x0, y0, DiaVars->FilmStripWidth, DiaVars->Height,
	DiaVars->FilmStripWidth, DiaVars->FilmStripHeight, 1);
}

/**
**	Draw horizontal film-strip.
**
**	@param x0	film-strip x start coordinate
**	@param y0	film-strip y start coordinate
*/
static void DiaDrawHorizontalFilmStrip(int x0, int y0)
{
    DiaDrawPage(x0, y0, DiaVars->Width, DiaVars->FilmStripHeight,
	DiaVars->FilmStripWidth, DiaVars->FilmStripHeight, 1);
}

/**
**	Draw single image.
**
**	Single image with or without film-strip.
*/
static void DiaDrawSingle(void)
{
    int x;
    int y;
    const FileEntry *entry;
    const char *file;

    x = 0;
    y = 0;
    if (DiaVars->FilmStrip) {
	// FIXME: no need to redraw film-strip, if already in backstore
	if (DiaVars->FilmStripHorizontal) {
	    DiaDrawHorizontalFilmStrip(x, y);
	    y = DiaVars->FilmStripHeight;
	} else {
	    DiaDrawVerticalFilmStrip(x, y);
	    x = DiaVars->FilmStripWidth;
	}
    }
    entry = EntryGet(DiaVars->FilesInDir, DiaVars->CurrentIndex);
    if (entry) {
	int ox;
	int oy;
	int w;
	int h;

	file = EntryName(entry);
	// FIXME: set clipping
	w = DiaVars->Width - x;
	h = DiaVars->Height - y;
	ox = DiaVars->OffsetX;
	oy = DiaVars->OffsetY;
	DiaShowImage(file, x + ox, y + oy, w - ox, h - oy);

	if (DiaVars->IndexLabel) {
	    DrawText(x + 2, y + 2, file, DiaVars->Color->Shadow, 0x0UL, w - 2,
		0, 4 + 1);
	    DrawText(x + 1, y + 1, file, DiaVars->Color->Text, 0x0UL, h - 3, 0,
		4 + 1);
	}

    }
}

/**
**	Draw double page image.
**
**	Double page images with or without film-strip.
*/
static void DiaDrawDouble(void)
{
    int x;
    int y;
    int index;
    const FileEntry *entry;
    const char *file;

    x = 0;
    y = 0;
    if (DiaVars->FilmStrip) {
	// FIXME: no need to redraw film-strip, if already in backstore
	if (DiaVars->FilmStripHorizontal) {
	    DiaDrawHorizontalFilmStrip(x, y);
	    y = DiaVars->FilmStripHeight;
	} else {
	    DiaDrawVerticalFilmStrip(x, y);
	    x = DiaVars->FilmStripWidth;
	}
    }
    entry = EntryGet(DiaVars->FilesInDir, DiaVars->CurrentIndex);
    if (entry) {
	int ox;
	int oy;
	int w;
	int h;

	file = EntryName(entry);
	// FIXME: set clipping
	w = (DiaVars->Width - x) / 2;
	h = DiaVars->Height - y;
	ox = DiaVars->OffsetX;
	oy = DiaVars->OffsetY;
	DiaShowImage(file, x + ox, y + oy, w - ox, h - oy);

	if (DiaVars->IndexLabel) {
	    DrawText(x + 2, y + 2, file, DiaVars->Color->Shadow, 0x0UL, w - 2,
		0, 4 + 1);
	    DrawText(x + 1, y + 1, file, DiaVars->Color->Text, 0x0UL, h - 3, 0,
		4 + 1);
	}
    }

    index = DiaVars->CurrentIndex;
    if ( (entry = EntryNext(DiaVars->FilesInDir, &index))) {
	int ox;
	int oy;
	int w;
	int h;

	file = EntryName(entry);
	// FIXME: set clipping
	w = (DiaVars->Width - x) / 2;
	x += w;
	h = DiaVars->Height - y;
	ox = DiaVars->OffsetX;
	oy = DiaVars->OffsetY;
	DiaShowImage(file, x + ox, y + oy, w - ox, h - oy);

	if (DiaVars->IndexLabel) {
	    DrawText(x + 2, y + 2, file, DiaVars->Color->Shadow, 0x0UL, w - 2,
		0, 4 + 1);
	    DrawText(x + 1, y + 1, file, DiaVars->Color->Text, 0x0UL, h - 3, 0,
		4 + 1);
	}
    }
}

/**
**	Draw index-page.
*/
static void DiaDrawIndex(void)
{
    DiaDrawPage((DiaVars->Width % DiaVars->IndexWidth) / 2,
	(DiaVars->Height % DiaVars->IndexHeight) / 2, DiaVars->Width,
	DiaVars->Height, DiaVars->IndexWidth, DiaVars->IndexHeight, 1);
}

/**
**	Draw dia.
*/
void DiaDraw(void)
{
    DiaVars->NeedRedraw = 0;

    // clear with background
    DrawRectangle(0, 0, DiaVars->Width, DiaVars->Height,
	DiaVars->Color->Background);

    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    //DiaDrawSingle();
	    DiaDrawDouble();
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaDrawIndex();
	    break;
    }
}

/**
**	Initialize diashow.
*/
void DiaInit(void)
{
    //
    //	default values
    //
    DiaVars->CornerWidth = DIA_CORNER_WIDTH;
    DiaVars->CornerHeight = DIA_CORNER_HEIGHT;

    DiaVars->SlideShowDelay = 30;

    DiaVars->FilmStripWidth = DIA_FILM_STRIP_WIDTH;
    DiaVars->FilmStripHeight = DIA_FILM_STRIP_HEIGHT;
    DiaVars->FilmStripScale = 100;

    DiaVars->IndexWidth = DIA_INDEX_WIDTH;
    DiaVars->IndexHeight = DIA_INDEX_HEIGHT;

    DiaVars->AspectNum = DIA_FILM_STRIP_WIDTH;
    DiaVars->AspectDen = DIA_FILM_STRIP_HEIGHT;

    DiaVars->SlideShow = 0;
    DiaVars->FilmStrip = 1;
    DiaVars->FilmStripHorizontal = 1;
    DiaVars->IndexLabel = 0;
    DiaVars->Layout = DIA_LAYOUT_SINGLE;

    DiaVars->Rotate = RotateNone;
    DiaVars->ScaleMethod = ScaleLanczos;

    DiaVars->Color->Background = 0x00000000;
    DiaVars->Color->Highlight = 0xFFFFFFFF;
    DiaVars->Color->Shadow = 0xFF000000;
    DiaVars->Color->Text = 0xFFFFFFFF;

    DiaVars->DoneInit = 1;
}

/**
**	Parse diashow config paramters.
**
**	@param name	paramter name (case sensetive)
**	@param value	value as string
**
**	@returns true if the parameter is supported.
*/
int DiaConfigParse(const char *name, const char *value)
{
    if (!DiaVars->DoneInit) {
	DiaInit();
    }

    fprintf(stderr, "dia: %s = %s\n", name, value);

    if (!strcasecmp(name, "Label")) {
	DiaVars->IndexLabel = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "FilmStrip")) {
	DiaVars->FilmStrip = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "FilmStripHorizontal")) {
	DiaVars->FilmStripHorizontal = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "SlideShow")) {
	DiaVars->SlideShow = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "Delay")) {
	DiaVars->SlideShowDelay = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "Backdrop")) {
	DiaVars->Backdrop = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "Fullscreen")) {
	DiaVars->Fullscreen = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "AspectNum")) {
	DiaVars->AspectNum = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "AspectDen")) {
	DiaVars->AspectDen = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "FilmStripWidth")) {
	DiaVars->FilmStripWidth = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "FilmStripHeight")) {
	DiaVars->FilmStripHeight = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "IndexWidth")) {
	DiaVars->IndexWidth = atoi(value);
	return 1;
    }
    if (!strcasecmp(name, "IndexHeight")) {
	DiaVars->IndexHeight = atoi(value);
	return 1;
    }

    if (!strcasecmp(name, "ScaleMethod")) {
	DiaVars->ScaleMethod = atoi(value);
	return 1;
    }

    if (!strcasecmp(name, "Color.Highlight")) {
	DiaVars->Color->Highlight = atoi(value);
	return 1;
    }

    return 0;
}

/**
**	Create the diashow window.
**
**	@param name	start file or directory.
**	@param width	diashow width
**	@param width	diashow height
*/
void DiaStart(const char *name, unsigned width, unsigned height)
{
    char *basename;
    const FileEntry *entry;
    int i;

    if (!DiaVars->DoneInit) {
	DiaInit();
    }
    DiaVars->Width = width;
    DiaVars->Height = height;

    DiaVars->SlideShowTick = GetMsTicks() + DiaVars->SlideShowDelay * 1000;

    free(DiaVars->Path);
    //
    //	directory or file, split path, basename
    //
    basename = strrchr(name, '/');
    if (basename) {
	i = ++basename + 1 - name;
	if ((DiaVars->Path = malloc(i))) {
	    memcpy(DiaVars->Path, name, i);
	    DiaVars->Path[i - 1] = 0;
	}
    } else {
	DiaVars->Path = strdup(name);
    }
    if (!DiaVars->Path) {
	Error(_("dia: out of memory\n"));
	return;
    }

    FileListDel(DiaVars->FilesInDir);
    DiaVars->FilesInDir = FileListNew(DiaVars->Path);
    // FIXME: handle error

    DiaVars->FirstIndex = 0;
    DiaVars->CurrentIndex = 0;
    // search basename as initial dia image
    if (basename && *basename) {
	i = 0;
	while ((entry = EntryNext(DiaVars->FilesInDir, &i))) {
	    if (!strcmp(basename, EntryName(entry))) {
		DiaVars->FirstIndex = i;
		// FIXME: show in the middle
		DiaVars->CurrentIndex = i;
		break;
	    }
	}
    }
}

/**
**	Destroy the diashow window.
**
*/
void DiaStop(void)
{
    FileListDel(DiaVars->FilesInDir);
    DiaVars->FilesInDir = NULL;
    free(DiaVars->Path);
    DiaVars->Path = NULL;
}

/**
**	Center film-strip or index-page.
*/
static void DiaCenterIndex(void)
{
    int i;
    int n;

    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    if (DiaVars->FilmStripHorizontal) {
		i = DiaVars->Width / DiaVars->FilmStripWidth;
	    } else {
		i = DiaVars->Height / DiaVars->FilmStripHeight;
	    }
	    // FIXME: multicol film-strip
	    n = i * 1;
	    break;
	case DIA_LAYOUT_INDEX:
	    i = DiaVars->Width / DiaVars->IndexWidth;
	    n = i * (DiaVars->Height / DiaVars->IndexHeight);
	    break;
    }
    // FIXME: try align the rows
    if (DiaVars->CurrentIndex < DiaVars->FirstIndex
	|| DiaVars->CurrentIndex >= DiaVars->FirstIndex + n) {
	DiaVars->FirstIndex = DiaVars->CurrentIndex - n / 2;
	if (DiaVars->FirstIndex < 0) {
	    DiaVars->FirstIndex = 0;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//	Commands
//////////////////////////////////////////////////////////////////////////////

/**
**	Previous dia image.
**
**	Keep film-strip upto date, show 1 image future.
**
**	@retval true	image changed
**	@retval false	image keeped
*/
static int DiaPrevScrollFilmStrip(void)
{
    int index;

    index = DiaVars->CurrentIndex;
    if (EntryPrev(DiaVars->FilesInDir, &index)) {
	DiaVars->CurrentIndex = index;
	// scroll film-strip 1 image backward
	if (DiaVars->CurrentIndex <= DiaVars->FirstIndex) {
	    // move start of index
	    index = DiaVars->FirstIndex;
	    if (EntryPrev(DiaVars->FilesInDir, &index)) {
		DiaVars->FirstIndex = index;
	    }
	}
	DiaDraw();
	return 1;
    }
    return 0;
}

/**
**	Previous dia image.
**
**	Keep index page upto date.
**
**	@retval true	image changed
**	@retval false	image keeped
*/
static int DiaPrevScrollIndex(void)
{
    int index;

    index = DiaVars->CurrentIndex;
    if (EntryPrev(DiaVars->FilesInDir, &index)) {
	DiaVars->CurrentIndex = index;
	// scroll index 1 row
	if (DiaVars->CurrentIndex < DiaVars->FirstIndex) {
	    int n;

	    // move start of index
	    index = DiaVars->FirstIndex;
	    n = DiaVars->Width / DiaVars->IndexWidth;
	    while (n-- && EntryPrev(DiaVars->FilesInDir, &index)) {
	    }
	    DiaVars->FirstIndex = index;
	}
	DiaDraw();
	return 1;
    }
    return 0;
}

/**
**	Next dia image.
**
**	Keep film-strip upto date, show 1 image future.
**
**	@retval true	image changed
**	@retval false	image keeped
*/
static int DiaNextScrollFilmStrip(void)
{
    int index;

    index = DiaVars->CurrentIndex;
    if (EntryNext(DiaVars->FilesInDir, &index)) {
	int n;

	DiaVars->CurrentIndex = index;
	if (DiaVars->FilmStripHorizontal) {
	    n = DiaVars->Width / DiaVars->FilmStripWidth;
	} else {
	    n = DiaVars->Height / DiaVars->FilmStripHeight;
	}
	// scroll film-strip 1 image forward
	if (DiaVars->CurrentIndex >= DiaVars->FirstIndex + n - 1) {
	    index = DiaVars->FirstIndex + n - 1;
	    // look how far we can move the end
	    if (EntryNext(DiaVars->FilesInDir, &index)) {
		// move start of index
		index = DiaVars->FirstIndex;
		if (EntryNext(DiaVars->FilesInDir, &index)) {
		    DiaVars->FirstIndex = index;
		}
	    }
	}
	DiaDraw();
	return 1;
    }
    return 0;
}

/**
**	Next dia image.
**
**	Keep index page upto date.
**
**	@retval true	image changed
**	@retval false	image keeped
*/
static int DiaNextScrollIndex(void)
{
    int index;

    index = DiaVars->CurrentIndex;
    if (EntryNext(DiaVars->FilesInDir, &index)) {
	int n;

	DiaVars->CurrentIndex = index;
	n = DiaVars->Width / DiaVars->IndexWidth;
	n *= DiaVars->Height / DiaVars->IndexHeight;

	// scroll index 1 row
	if (DiaVars->CurrentIndex >= DiaVars->FirstIndex + n) {
	    int i;

	    index = DiaVars->FirstIndex + n - 1;
	    n = DiaVars->Width / DiaVars->IndexWidth;
	    // look how far we can move the end
	    for (i = 0; i < n; ++i) {
		if (!EntryNext(DiaVars->FilesInDir, &index)) {
		    break;
		}
	    }

	    // move start of index
	    index = DiaVars->FirstIndex;
	    while (i-- && EntryNext(DiaVars->FilesInDir, &index)) {
	    }
	    DiaVars->FirstIndex = index;
	}
	DiaDraw();
	return 1;
    }
    return 0;
}

/**
**	Scroll film-strip up.
*/
static void DiaScrollFilmUp(void)
{
    int index;

    printf("%d . %d\n", DiaVars->CurrentIndex, DiaVars->FirstIndex);
    // move start of index
    index = DiaVars->FirstIndex;
    if (EntryPrev(DiaVars->FilesInDir, &index)) {
	DiaVars->FirstIndex = index;
    }
    index = DiaVars->CurrentIndex;
    if (EntryPrev(DiaVars->FilesInDir, &index)) {
	DiaVars->CurrentIndex = index;
    }
    printf("->%d . %d\n", DiaVars->CurrentIndex, DiaVars->FirstIndex);
    DiaDraw();
}

/**
**	Scroll film-strip down.
*/
static void DiaScrollFilmDown(void)
{
    int index;
    int n;

    printf("%d . %d\n", DiaVars->CurrentIndex, DiaVars->FirstIndex);
    // move end of index
    if (DiaVars->FilmStripHorizontal) {
	n = DiaVars->Width / DiaVars->FilmStripWidth;
    } else {
	n = DiaVars->Height / DiaVars->FilmStripHeight;
    }
    index = DiaVars->FirstIndex + n - 1;
    if (EntryNext(DiaVars->FilesInDir, &index)) {
	index = DiaVars->FirstIndex;
	EntryNext(DiaVars->FilesInDir, &index);
	DiaVars->FirstIndex = index;
    }
    index = DiaVars->CurrentIndex;
    if (EntryNext(DiaVars->FilesInDir, &index)) {
	DiaVars->CurrentIndex = index;
    }
    printf("->%d . %d\n", DiaVars->CurrentIndex, DiaVars->FirstIndex);
    DiaDraw();
}

/**
**	Move page up.
**
**	@param n	number of images to page
*/
static void DiaPageUp(int n)
{
    int i;
    int index;

    // look how far we can move the first
    index = DiaVars->FirstIndex;
    for (i = 0; i < n; ++i) {
	if (!EntryPrev(DiaVars->FilesInDir, &index)) {
	    break;
	}
    }
    DiaVars->FirstIndex = index;
    // move current
    index = DiaVars->CurrentIndex;
    n = i;
    for (i = 0; i < n; ++i) {
	if (!EntryPrev(DiaVars->FilesInDir, &index)) {
	    // shouldn't happen
	    break;
	}
    }
    DiaVars->CurrentIndex = index;

    DiaDraw();
}

/**
**	Move page down.
**
**	@param n	number of images to page
**	@param e	page end
*/
static void DiaPageDown(int n, int e)
{
    int i;
    int index;

    // look how far we can move the end
    index = DiaVars->FirstIndex + e - 1;
    for (i = 0; i < n; ++i) {
	if (!EntryNext(DiaVars->FilesInDir, &index)) {
	    break;
	}
    }
    // move start of index
    index = DiaVars->FirstIndex;
    n = i;
    for (i = 0; i < n; ++i) {
	if (!EntryNext(DiaVars->FilesInDir, &index)) {
	    // should not happen
	    break;
	}
    }
    DiaVars->FirstIndex = index;
    // move current
    index = DiaVars->CurrentIndex;
    for (i = 0; i < n; ++i) {
	if (!EntryNext(DiaVars->FilesInDir, &index)) {
	    // should not happen
	    break;
	}
    }
    DiaVars->CurrentIndex = index;

    DiaDraw();
}

/**
**	Page film-strip 1 page up.
*/
static void DiaPageFilmStripUp(void)
{
    if (DiaVars->FilmStripHorizontal) {
	DiaPageUp(DiaVars->Width / DiaVars->FilmStripWidth);
    } else {
	DiaPageUp(DiaVars->Height / DiaVars->FilmStripHeight);
    }
}

/**
**	Page film-strip 1 page down.
*/
static void DiaPageFilmStripDown(void)
{
    int n;

    if (DiaVars->FilmStripHorizontal) {
	n = DiaVars->Width / DiaVars->FilmStripWidth;
    } else {
	n = DiaVars->Height / DiaVars->FilmStripHeight;
    }
    DiaPageDown(n, n);
}

/**
**	Page index-page 1 page up.
*/
static void DiaPageIndexPageUp(void)
{
    DiaPageUp((DiaVars->Height / DiaVars->IndexHeight) * (DiaVars->Width /
	    DiaVars->IndexWidth));
}

/**
**	Page index-page 1 page down.
*/
static void DiaPageIndexPageDown(void)
{
    int n;

    n = (DiaVars->Height / DiaVars->IndexHeight) * (DiaVars->Width /
	DiaVars->IndexWidth);
    DiaPageDown(n, n);
}

/**
**	Dia go index page up.
**
**	Keep index-page upto date.
*/
static void DiaUpIndex(void)
{
    int i;
    int n;
    int index;

    // move cursor 1 row up
    index = DiaVars->CurrentIndex;
    n = DiaVars->Width / DiaVars->IndexWidth;
    for (i = n; i; i--) {
	if (!EntryPrev(DiaVars->FilesInDir, &index)) {
	    return;
	}
    }
    DiaVars->CurrentIndex = index;

    if (DiaVars->CurrentIndex < DiaVars->FirstIndex) {
	// move start of index
	index = DiaVars->FirstIndex;
	while (n--) {
	    if (!EntryPrev(DiaVars->FilesInDir, &index)) {
		break;
	    }
	}
	DiaVars->FirstIndex = index;
    }
    DiaDraw();
}

/**
**	Dia go index page up.
**
**	Keep index-page upto date.
**	Next dia image.
*/
static void DiaDownIndex(void)
{
    int i;
    int n;
    int index;

    // move cursor 1 row down
    index = DiaVars->CurrentIndex;
    n = DiaVars->Width / DiaVars->IndexWidth;
    for (i = n; i; i--) {
	if (!EntryNext(DiaVars->FilesInDir, &index)) {
	    return;
	}
    }
    DiaVars->CurrentIndex = index;

    i = n;
    n *= DiaVars->Height / DiaVars->IndexHeight;

    // scroll index 1 row
    if (DiaVars->CurrentIndex >= DiaVars->FirstIndex + n) {
	index = DiaVars->FirstIndex + n - 1;

	n = i;
	// look how far we can move the end
	for (i = 0; i < n; ++i) {
	    if (!EntryNext(DiaVars->FilesInDir, &index)) {
		break;
	    }
	}

	// move start of index
	index = DiaVars->FirstIndex;
	while (i--) {
	    if (!EntryNext(DiaVars->FilesInDir, &index)) {
		break;
	    }
	}
	DiaVars->FirstIndex = index;
    }
    DiaDraw();
}

#if 0

i = DiaVars->Width / DiaVars->IndexWidth;
i *= DiaVars->Height / DiaVars->IndexHeight;

/**
**	Dia previous index-page.
*/
static void DiaPrevPage(void)
{
    int n;

    n = DiaVars->Width / DiaVars->IndexWidth;
    n *= DiaVars->Height / DiaVars->IndexHeight;

    DiaIndexScrollUp(n);
}

/**
**	Dia next index-page.
*/
static void DiaNextPage(void)
{
    int n;

    n = DiaVars->Width / DiaVars->IndexWidth;
    n *= DiaVars->Height / DiaVars->IndexHeight;

    DiaIndexScrollDown(n);
}

#endif

//////////////////////////////////////////////////////////////////////////////
//	Commands

/**
**	Dia up key command.
*/
void DiaCmdUp(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaPrevScrollFilmStrip();
	    // FIXME: reached first dia image, change path
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaUpIndex();
	    break;
    }
}

/**
**	Dia down key command.
*/
void DiaCmdDown(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaNextScrollFilmStrip();
	    // FIXME: reached last dia image, change path
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaDownIndex();
	    break;
    }
}

/**
**	Dia left key command.
*/
void DiaCmdLeft(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaPageFilmStripUp();
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaPrevScrollIndex();
	    break;
    }
}

/**
**	Dia right key command.
*/
void DiaCmdRight(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaPageFilmStripDown();
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaNextScrollIndex();
	    break;
    }
}

/**
**	Dia page up.
*/
void DiaCmdPageUp(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaScrollFilmUp();
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaPageIndexPageUp();
	    break;
    }
}

/**
**	Dia page down.
*/
void DiaCmdPageDown(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaScrollFilmDown();
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaPageIndexPageDown();
	    break;
    }
}

/**
**	Enter command.
*/
void DiaCmdEnter(void)
{
    // FIXME: toggle layout, planned Index -> Single -> Zoom
    DiaCmdRotateLayout();
}

/**
**	Rotate layout.
*/
void DiaCmdRotateLayout(void)
{
    DiaVars->Layout++;
    DiaCenterIndex();
    DiaDraw();
}

/**
**	Toggle film-strip on/off.
*/
void DiaCmdToggleFilmStrip(void)
{
    DiaVars->FilmStrip++;
    DiaDraw();
}

/**
**	Rotate zoom mode.	
*/
void DiaCmdRotateZoom(void)
{
    // FIXME: write zoom modes
}

/**
**	Rotate image 90 counterclockwise.
*/
void DiaCmdRotateCounterClockwise(void)
{
}

/**
**	Rotate image 90 clockwise.
*/
void DiaCmdRotateClockwise(void)
{
}

/**
**	Shrink thumbnails.
*/
void DiaCmdShrink(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    if (DiaVars->FilmStripWidth > 2) {
		DiaVars->FilmStripWidth--;
		DiaVars->FilmStripHeight =
		    (DiaVars->FilmStripWidth * DiaVars->AspectDen) /
		    DiaVars->AspectNum;
	    }
	    break;
	case DIA_LAYOUT_INDEX:
	    if (DiaVars->IndexWidth > 2) {
		DiaVars->IndexWidth--;
		DiaVars->IndexHeight =
		    (DiaVars->IndexWidth * DiaVars->AspectDen) /
		    DiaVars->AspectNum;
	    }
	    break;
    }

    DiaCenterIndex();
    DiaDraw();
}

/**
**	Grow thumbnails.
*/
void DiaCmdGrow(void)
{
    switch (DiaVars->Layout) {
	case DIA_LAYOUT_SINGLE:
	    DiaVars->FilmStripWidth++;
	    DiaVars->FilmStripHeight =
		(DiaVars->FilmStripWidth * DiaVars->AspectDen) /
		DiaVars->AspectNum;
	    break;
	case DIA_LAYOUT_INDEX:
	    DiaVars->IndexWidth++;
	    DiaVars->IndexHeight =
		(DiaVars->IndexWidth * DiaVars->AspectDen) /
		DiaVars->AspectNum;
	    break;
    }

    DiaCenterIndex();
    DiaDraw();
}

/**
**	Show image info.
*/
void DiaCmdInfo(void)
{
}

/**
**	Start/Stop dia-show slide-show.
*/
void DiaCmdToggleSlideShow(void)
{
}
